/* Implementation of RT-Linux kernel module rt_das08jr.
 *
 * Real-Time Linux driver for the ComputerBoards CIO-DAS08/Jr-AO.
 *
 * Copyright (C) 1998-2000 Jochen Kpper
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file License. if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * $Id: rt_das08jr.c,v 1.7 2000/05/14 11:27:11 jochen Exp $ */


#define __KERNEL__
#define __RT__
#define MODULE


#include <linux/config.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>

#include <asm/io.h>

#ifdef RTAI
#  include <rtai.h>
#  include <rtai_sched.h>
#else
#endif

#include "rt_das08jr.h"


/* Module Name */
#define RT_DAS08JR_NAME "rt_das08jr"


/* driver version */
static const char rt_das08jr_version[] = VERSION;


/* user configurable port base
 * If you give a parameter io=0xNNN, this address will be used,
 * autoprobing will be done otherwise. */
unsigned long io = 0;

MODULE_AUTHOR( "Jochen Kuepper" );
MODULE_PARM( io, "l" );

/*
EXPORT_SYMBOL( rt_das08jr_close );
EXPORT_SYMBOL( rt_das08jr_open );
EXPORT_SYMBOL( rt_das08jr_read );
EXPORT_SYMBOL( rt_das08jr_write );
*/

/* port addresses to autoprobe
 * If the parameter io=0xNNN is not given at load time, these addresses are
 * probed in their respective order until one device is found or all specified
 * addresses are probed */
static unsigned long autoprobe_port[] =
{
	0x300,
	0x310,
	0x320,
	0x200,
	0x270,
	0x2e0,
	0x2e8,
	0x2f0,
	0x2f8,
	0x378,
	0
};



/* number of available lines */
#define RT_DAS08JR_ANALOG_INPUTS    ( 8 )
#define RT_DAS08JR_ANALOG_OUTPUTS   ( 2 )
#define RT_DAS08JR_DIGITAL_LINES    ( 1 )
#define RT_DAS08JR_MAX_LINE         ( 8 ) /* we have nine lines starting at zero */

/* number of ports */
#define RT_DAS08JR_SIZE             ( 8 )

/* Port Offsets */
#define RT_DAS08JR_ANALOG_IN_LSB    ( 0 ) /* read */
#define RT_DAS08JR_ANALOG_IN_MSB    ( 1 ) /* read */
#define RT_DAS08JR_START            ( 1 )
#define RT_DAS08JR_STATUS           ( 2 ) /* read */
#define RT_DAS08JR_MUX              ( 2 ) /* read */
#define RT_DAS08JR_CONTROL          ( 2 )
#define RT_DAS08JR_DIGITAL_IN       ( 3 ) /* read */
#define RT_DAS08JR_DIGITAL_OUT      ( 3 )
#define RT_DAS08JR_ANALOG_OUT_LSB0  ( 4 )
#define RT_DAS08JR_ANALOG_OUT_MSB0  ( 5 )
#define RT_DAS08JR_ANALOG_OUT_LSB1  ( 6 )
#define RT_DAS08JR_ANALOG_OUT_MSB1  ( 7 )



/** all information about a single channel */
#define RT_DAS08JR_ANALOG_READ      0x01
#define RT_DAS08JR_ANALOG_WRITE     0x02
#define RT_DAS08JR_DIGITAL_READ     0x04
#define RT_DAS08JR_DIGITAL_WRITE    0x04

struct channel 
{
    int magic;
    int open;
    int type;
};

/** device description
 *
 * You might neeed to adopt lines two and three to your card (AO or not).  */
struct channel rt_das08jr[ RT_DAS08JR_MAX_LINE + 1 ] = {
    { 0, 0, RT_DAS08JR_DIGITAL_READ | RT_DAS08JR_DIGITAL_WRITE },
    { 1, 0, RT_DAS08JR_ANALOG_READ | RT_DAS08JR_ANALOG_WRITE },
    { 2, 0, RT_DAS08JR_ANALOG_READ | RT_DAS08JR_ANALOG_WRITE },
    { 3, 0, RT_DAS08JR_ANALOG_READ },
    { 4, 0, RT_DAS08JR_ANALOG_READ },
    { 5, 0, RT_DAS08JR_ANALOG_READ },
    { 6, 0, RT_DAS08JR_ANALOG_READ },
    { 7, 0, RT_DAS08JR_ANALOG_READ },
    { 8, 0, RT_DAS08JR_ANALOG_READ }
};



/** Check for a free device at the given port base address
 *
 * @param base Port address to check for a device at.
 * @return     0 if there is a free card at that address,
 *             negative error otherwise
 *
 * @author Jochen Kpper
 * @version 2000/03/15 */
static int das08jr_check_port( unsigned long base )
{
    /* check and register given port region */
    if( -EBUSY == check_region( base, RT_DAS08JR_SIZE ) )
	return( -EBUSY );
    request_region( base, RT_DAS08JR_SIZE, RT_DAS08JR_NAME );
    
    /* Try out the port */
    outb_p( 0x03, base + RT_DAS08JR_CONTROL );
    if( 0x03 != ( 0x07 & inb_p( base + RT_DAS08JR_MUX ) ) ) {
	printk( KERN_DEBUG RT_DAS08JR_NAME ": no device at %lx\n", base );
	release_region( base, RT_DAS08JR_SIZE );
	return( -ENODEV );
    }
    
    outb_p( 0x06, base + RT_DAS08JR_CONTROL );
    if( 0x06 != ( 0x07 & inb_p( base + RT_DAS08JR_MUX ) ) ) {
	printk( KERN_DEBUG RT_DAS08JR_NAME ": no device at %lx\n", base );
	release_region( base, RT_DAS08JR_SIZE );
	return( -ENODEV );
    }
    
    return( 0 );
}


/** Initialization
 *
 * Probe for device, register ports, register as character device,
 * export exactly the symbols we want to be public (none yet).
 *
 * @return Success status, zero on success.
 *
 * @author Jochen Kpper
 * @version 2000/03/15 */
int init_module( void )
{
    /* check device port base and request the port region */
    if( 0 == io ) {
	/* autoprobe */
	int i = 0;
	while ( 0 != ( io = autoprobe_port[ i ] ) ) {
	    if( 0 >= das08jr_check_port( io ) )
		break;
	    i++;
	}
	if( 0 == io ) {
	    return( -ENODEV );
	}
    } else {
	/* use specified base */
	int error;
	if( 0 >= ( error = das08jr_check_port( io ) ) )
	    return( error );
    }
    
    /* finish up */
    printk( KERN_INFO RT_DAS08JR_NAME ": CIO-DAS08Jr-AO real-time driver (version %s) loaded\n"
	    KERN_INFO RT_DAS08JR_NAME ": Copyright (C) 1998-2000 Jochen Kpper\n"
	    KERN_INFO RT_DAS08JR_NAME ": using board at 0x%lx\n",
	    rt_das08jr_version, io );
    return( 0 );
}



/** Cleanup when we are unloaded.
 *
 * @author Jochen Kpper
 * @version 2000/03/15 */
void cleanup_module( void )
{
    release_region( io, RT_DAS08JR_SIZE );
    printk( KERN_INFO RT_DAS08JR_NAME ": unloaded\n" );
}



/** Close
 *
 * @author Jochen Kpper
 * @version 2000/03/15 */
extern int rt_das08jr_close( int channel )
{
    if( 0 == rt_das08jr[ channel ].open ) {
	return( -EPERM );
    } else {
	rt_das08jr[ channel ].open = 0;
	return( 0 );
    }
}



/** Open
 *
 * @author Jochen Kpper
 * @version 2000/03/15 */
int rt_das08jr_open( int channel )
{
    if( 0 != rt_das08jr[ channel ].open ) {
	return( -EBUSY );
    } else {
	rt_das08jr[ channel ].open = 1;
	return( 0 );
    }
}




/** Read from analog line
 *
 * We always read one short value from the channel specified.
 *
 * @param channel
 * @param value
 * @return         0 on success, negative error condition otherwise
 *
 * @author Jochen Kpper
 * @version 2000/05/05 */
int rt_das08jr_read( int channel, unsigned short int *value )
{
    if( 0 >= rt_das08jr[ channel ].open )
	return( -EPERM );
    else if( ! (RT_DAS08JR_ANALOG_READ & rt_das08jr[ channel ].type) )
	return( -EPERM );	
    else {
	unsigned short int lsb, msb;
	
	/* select channel and trigger conversion */
	outb_p( channel - 1, io + RT_DAS08JR_CONTROL );
	outb_p( 0xff, io + RT_DAS08JR_START );

	/* Wait for conversion to finish (loop until EOC clear) --- we
	 * first give away the cpu for a typical conversion time, and
	 * busy-wait until the divice is ready afterwards.
	 *
	 * Note that rt_sleep might take considerably longer than the
	 * specified 25 us - probably as much as twice as long on a
	 * PII. If you want quick returns, better use
	 * rt_busy_sleep. */
#if (defined RTAI) && (BUSYWAIT == 0)
	rt_sleep( nano2count( 25000 ) );
#endif
	while( 0x80 & inb_p( io + RT_DAS08JR_STATUS ) )
	    ;
	
	/* read data and put it into user space */
	lsb = inb_p( io + RT_DAS08JR_ANALOG_IN_LSB );
	msb = inb_p( io + RT_DAS08JR_ANALOG_IN_MSB );
	*value = (msb << 4) + (lsb >> 4);
	return( 0 );
    }
}



/** Write to output line
 *
 * We always read one short value from the channel specified.
 *
 * @param channel
 * @param value
 * @return         0 on success, negative error condition otherwise
 *
 * @author Jochen Kpper
 * @version 2000/05/04 */
int rt_das08jr_write( int channel, unsigned short int value )
{
    if( 0 >= rt_das08jr[ channel ].open )
	return( -EPERM );
    else if( ! ( ( RT_DAS08JR_ANALOG_WRITE & rt_das08jr[ channel ].type )
		 || ( (RT_DAS08JR_DIGITAL_WRITE & rt_das08jr[ channel ].type) ) ) )
	return( -EPERM );
    else {
	if( 0 == channel ) {
	    outb( (unsigned char)value, io + RT_DAS08JR_DIGITAL_OUT );
	} else if( 1 == channel ) {
	    outb( (unsigned char)( value >> 8 ), io + RT_DAS08JR_ANALOG_OUT_MSB1 );
	    outb( (unsigned char)( value & 0xff ), io + RT_DAS08JR_ANALOG_OUT_LSB1 );
	} else {
	    outb( (unsigned char)( value >> 8 ), io + RT_DAS08JR_ANALOG_OUT_MSB0 );
	    outb( (unsigned char)( value & 0xff ), io + RT_DAS08JR_ANALOG_OUT_LSB0 );
	}
	/* Read DIO control register to initiate DAC */
	inb_p( io + RT_DAS08JR_DIGITAL_IN);
	return( 0 );
    }
}



/**
 * Local Variables:
 * mode: C
 * c-file-style: "Stroustrup"
 * End:
 */
